# -*- coding: utf-8 -*-
import logging
import sys
import zmq

from functools import partial
from zmq.eventloop import ioloop
from collections import namedtuple

from .input import StreamInput
from .output import StreamOutput
from .utils import index_by_dict_key, index_by_obj_prop
from .config import load_config_file

logger = logging.getLogger(__name__)

ioloop.install()


class StreamRouter(object):
    """
    Router instance. Handles input and output connections & routing (priority).
    """
    # `_source_ports` and `_destinations` are only used for initial router setup (during instance `start`)
    _source_ports = []
    _destinations = []
    _inputs = []
    _current_input = None
    _last_input = None
    _forced_port = None
    _outputs = []
    _current_config = None
    _state = None
    _start_with_ioloop = []
    _state_change_callbacks = []

    DelaySettings = namedtuple('DelaySettings', [
        'nodata_failover_delay',
        'nodata_recover_delay',
        'noaudio_failover_delay',
        'noaudio_recover_delay',
        'audio_threshold'])

    def __init__(self, config, **options):
        self._current_config = config

        self._source_ports = config.get('source_ports')
        self._destinations = config.get('destinations')

        self.delay_settings = StreamRouter.DelaySettings(
                config.get('nodata_failover_delay'),
                config.get('nodata_recover_delay'),
                config.get('noaudio_failover_delay'),
                config.get('noaudio_recover_delay'),
                config.get('audio_threshold'))

        self.zmq_ctx = zmq.Context()

    def get_current_input(self):
        return self._current_input

    def set_auto_switch(self):
        self._forced_port = None

    def force_input(self, i):
        if i not in [inp.port for inp in self._inputs]:
            raise ValueError("Port {} not available".format(i))
        self._forced_port = i





    ###################################################################
    # input a.k.a. source handling
    ###################################################################
    def add_input(self, port, before=False):
        i = StreamInput(self.zmq_ctx, port, self.delay_settings)
        if before:
            self._inputs.insert(0, i)
        else:
            self._inputs.append(i)
        logger.info('Created input socket on port {}'.format(port))


    def remove_input(self, port):
        index = index_by_obj_prop(self._inputs, 'port', port)
        if index > -1:
            i = self._inputs.pop(index)
            i.stop()
        logger.info('Removed input socket on port {}'.format(port))

    def get_inputs(self):
        return self._inputs

    def get_input_ports(self):
        return [i.port for i in self._inputs]





    ###################################################################
    # output a.k.a. destination handling
    ###################################################################
    def add_output(self, destination):
        o = StreamOutput(self.zmq_ctx, destination)
        logger.info('Connected output to {}'.format(destination))
        self._outputs.append(o)

    def remove_output(self, destination):
        index = index_by_obj_prop(self._outputs, 'output', destination)
        if index > -1:
            o = self._outputs.pop(index)
            o.stop()
        logger.info('Removed output to {}'.format(destination))

    def get_outputs(self):
        return self._outputs

    def get_output_destinations(self):
        return [o.output for o in self._outputs]


    def set_current_input(self):
        current_input = None

        if not self._inputs:
            self._current_input = None
            return

        # force input
        if self._forced_port:
            filtered_inputs = [i for i in self._inputs if i.port == self._forced_port]
            if filtered_inputs:
                current_input = filtered_inputs[0]

        # Loop through the inputs and return the first available one.
        if not current_input:
            available_inputs = [i for i in self._inputs if i.is_available]
            if available_inputs:
                current_input = available_inputs[0]

        self._current_input = current_input

        if current_input is not None and current_input != self._last_input:
            logger.info('Switching inputs: {} to {}'.format(self._last_input.port if self._last_input else None, current_input.port))
            self._last_input = current_input


    def route(self, stream, msg, input):
        """
        Routes the active input to all outputs
        """

        # Trigger a 'heartbeat' tick on the input.
        input.tick(msg[0])

        self.set_current_input()

        if input == self._current_input:
            for o in self._outputs:
                o.send(msg[0])

    def start(self):

        for port in self._source_ports:
            self.add_input(port=port)
        self._source_ports = []

        for destination in self._destinations:
            self.add_output(destination)
        self._destinations = []

        for i in self._inputs:
            i.stream.on_recv_stream(partial(self.route, input=i))

        for cb in self._start_with_ioloop:
            if not cb.started:
                logger.debug('Starting with ioloop: {}'.format(cb))
                cb.start()

        # Initialise the tornado ioloop
        # http://pyzmq.readthedocs.io/en/latest/eventloop.html#futures-and-coroutines
        io_loop = ioloop.IOLoop.instance()

        io_loop.start()

        while True:
            pass

    def stop(self):
        logger.info('Stopping router')
        for port in self.get_input_ports():
            self.remove_input(port=port)

        for destination in self.get_output_destinations():
            self.remove_output(destination)

        sys.exit()

    def register_start_with_ioloop(self, cb):
        if not cb in self._start_with_ioloop:
            self._start_with_ioloop.append(cb)

    def reload_configuration(self):
        config_file = self._current_config.get('config_file')
        config = load_config_file(config_file)

        self.delay_settings = StreamRouter.DelaySettings(
                config.get('nodata_failover_delay'),
                config.get('nodata_recover_delay'),
                config.get('noaudio_failover_delay'),
                config.get('noaudio_recover_delay'),
                config.get('audio_threshold'))

        ###############################################################
        # reconfigure outputs, a.k.a. 'destinations'
        ###############################################################
        current_destinations = self.get_output_destinations()
        new_destinations = config.get('destinations')

        # stop removed outputs
        for destionation in current_destinations:
            if not destionation in new_destinations:
                self.remove_output(destionation)

        # add newly configured outputs
        for destionation in new_destinations:
            if not destionation in current_destinations:
                self.add_output(destionation)



        ###############################################################
        # reconfigure inputs, a.k.a. 'source_ports'
        ###############################################################
        current_input_ports = self.get_input_ports()
        new_input_ports = config.get('source_ports')

        # stop removed inputs
        for port in current_input_ports:
            if not port in new_input_ports:
                self.remove_input(port)

        # add newly configured inputs
        for port in new_input_ports:
            if not port in current_input_ports:
                self.add_input(port)

        # update order / priority of inputs
        # TODO: 100% sure there is a better way to implement this :)
        _inputs = []
        for port in new_input_ports:
            for i in self._inputs:
                if i.port == port:
                    _inputs.append(i)

        # replace inputs with updated/sorted values
        self._inputs = _inputs




# -*- coding: utf-8 -*-

def index_by_dict_key(lst, key, value):
    for i, dic in enumerate(lst):
        if dic[key] == value:
            return i
    return -1

def index_by_obj_prop(list, prop, value):
    for i, obj in enumerate(list):
        if getattr(obj, prop) == value:
            return i
    return -1





# -*- coding: utf-8 -*-
from tornado import gen
from tornado.iostream import StreamClosedError
from tornado.netutil import bind_unix_socket
from tornado.tcpserver import TCPServer

from .executor import CommandExecutor

RC_PROMPT_PREFIX = b'> '

class RCServer(TCPServer, CommandExecutor):
    """
    exposes a control interface via telnet and/or unix-socket
    """
    router = None

    def __init__(self, router):
        self.router = router
        super(RCServer, self).__init__()


    def __str__(self):
        return '<RCServer>'

    @property
    def started(self):
        return self._started

    def add_unix_socket(self, socket):
        """
        helper method to add file-based socket by path.
        so we can avoid that `bind_unix_socket` has to be imported at the place the
        `RCServer` is initialised.
        """
        self.add_socket(bind_unix_socket(socket))

    @gen.coroutine
    def handle_stream(self, stream, address):
        """
        handle telnet connection
        http://www.tornadoweb.org/en/stable/gen.html#tornado-gen-simplify-asynchronous-code
        """
        #stream.write(RC_PROMPT_PREFIX)
        while True:
            try:
                command = yield stream.read_until(b'\n')
                try:
                    cmd = command.decode().strip()
                    if cmd == "quit":
                        stream.close()
                        break
                    result = self.handle_command(cmd)
                except TypeError as e:
                    result = 'Error: {}'.format(e)
                yield stream.write(result.encode() + RC_PROMPT_PREFIX)
            except StreamClosedError:
                break



# -*- coding: utf-8 -*-
import logging
import os
import sys
import click

from .config import DEFAULT_CONFIG, parse_config, load_config_file
from .router import StreamRouter
from .control.rcserver import RCServer

from odroute import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.option('-v', '--verbose', default=False, is_flag=True)
def odr_cli(verbose):
    if verbose:
        logger.setLevel(logging.DEBUG)

@odr_cli.command()
def version():
    click.echo(__version__)
    sys.exit()

@odr_cli.command()
@click.option('--config', '-c', 'config_file',
              # `type` is set to `str` and as we need to handle file reading manually (so not `click.File()`)
              type=str, required=False,
              help='Use the specified configuration file instead of CLI options'
              )
# sources (a.k.a. inputs) and outputs
@click.option('--source', '-s', 'source_ports',
              type=int, multiple=True, required=False,
              help='The source ports for incoming connections. Can (and likely will) be given multiple times'
              )
@click.option('--output', '-o', 'destinations',
              type=str, multiple=True, required=False,
              help='Destinations to route to, in the form of: tcp://<hostname>:<port>. Can be given multiple times'
              )
# no-audio and no-data failovers
@click.option('--delay', '-d', 'nodata_failover_delay',
              default=0.5,
              help='Delay for falling back to secondary streams in case of missing data'
              )
@click.option('--delay-recover', '-r', 'nodata_recover_delay',
              default=0.5,
              help='Delay for falling back to primary streams in case of recovery'
              )
@click.option('--audio-threshold', '-a', 'audio_threshold',
              default=-90,
              help='Minimum audio level (range [-90..0] dB) for input to be considered ok. Set to -90 to disable level detection'
              )
@click.option('--audio-delay', '-D', 'noaudio_failover_delay',
              default=10,
              help='Delay for falling back to secondary streams in case of audio level below threshold'
              )
@click.option('--audio-delay-recover', '-R', 'noaudio_recover_delay',
              default=10,
              help='Delay for falling back to primary streams in case of audio level recovery'
              )
# 'external' controls
@click.option('--telnet', '-t',
              required=False,
              help='Add telnet interface: <bind addr>:<port> or <port> (if only port is given interface will bind to 127.0.0.1)'
              )
@click.option('--socket', '-S',
              required=False,
              help='Add unix socket interface: </path/to/socket>'
              )
@click.option('--master', '-m',
              required=False,
              help='Connect odroute instance to master control-server: tcp://<hostnam>:<port>'
              )
def run(config_file, **config):

    #print(config)

    if config_file:
        config_file = os.path.abspath(config_file)
        click.echo('-' * 72)
        click.echo('Using config file - other CLI options will be ignored.\nConfig: {}'.format(config_file))
        click.echo('-' * 72)
        parsed_config = load_config_file(config_file)

    else:
        parsed_config = parse_config(config)


    # primitive logger configuration
    logging.basicConfig(
        format=parsed_config.get('log_format'),
        level=parsed_config.get('log_level'),
    )

    while True:
        r = StreamRouter(config=parsed_config)

        # adding remote-control interface
        telnet, socket = parsed_config.get('telnet', False), parsed_config.get('socket', False)
        if telnet or socket:
            rc_server = RCServer(router=r)

            if telnet:
                if telnet.isdigit():
                    _address, _port,  = '127.0.0.1', int(telnet)
                else:
                    _address, _port = telnet.split(':')

                logger.info('Binding telnet interface to {}:{}'.format(_address, _port))
                rc_server.bind(address=_address, port=int(_port))

            if socket:
                logger.info('Binding unix socket interface to {}'.format(socket))
                rc_server.add_unix_socket(socket)

            r.register_start_with_ioloop(rc_server)


        try:
            #r.connect()
            r.start()
        except KeyboardInterrupt:
            logger.info('Ctrl-c received. Stopping router.')
            r.stop()
            sys.exit()




# -*- coding: utf-8 -*-
import logging
import re
import time

COMMANDS = [
    'help',
    'input.list',
    'input.list (?:[a-z]+)',
    'input.current',
    'input.force (?:[0-9]+|auto)',
    'input.add (?:[0-9]+)',
    'input.remove (?:[0-9]+)',
    'output.list',
    'output.list (?:[a-z]+)',
    'output.add (?:[a-z0-9\:\/]+)',
    'output.remove (?:[a-z0-9\:\/]+)',
    'reload',
]

# the quit command is implemented in rcserver.py
COMMANDS_HELP = '''
# odroute rc interface
  commands:
  quit:                        disconnect the RC control
  help:                        returns available list of commands
  input.list [detail]          returns router inputs [<port>]
  input.current                returns active input [<port>]
  input.force <port>|auto      force input on <port>; returns forced|auto input [<port>]
  input.add <port>             add input on port <port>
  input.remove <port>          removes input on port <port>
  output.list [detail]         returns router outputs [<tcp://host:port>]
  output.add <destination>     adds output to <tcp://host:port>
  output.remove <destination>  removes output <tcp://host:port>
  reload                       reloads configuration (only available if configured via file)
'''

logger = logging.getLogger(__name__)

class CommandExecutor(object):
    """
    implements a generic interface to send commands to the router instance.
    currently used by telnet and socket interface, partially implemented by
    `RoutingMinion`.
    implementation should use the `handle_command` method (see below) - however
    if you have reasons you can access the `router` directly at own risk.
    """

    router = None

    def __init__(self, router):
        self.router = router

    def handle_command(self, command):
        """
        handles incoming commands & calls respective methods
        input:
        `foo.bar abc xzz`
        is mapped to:
        `self.cmd_foo_bar(*['abc', 'xyz'])`
        """
        pattern = '(' + ')|('.join(COMMANDS) + ')'
        if not re.match(pattern, command):
            return 'invalid command. try "help"\n'

        _c = command.split(' ')
        _cmd, _args = _c[0].replace('.', '_'), _c[1:]

        try:
            return getattr(self, 'cmd_' + _cmd)(*_args)
        except (AttributeError) as e:
            logger.warning('error executing command {} - {}'.format(_cmd, e))
            return '{}'.format(e)

    ###################################################################
    # commands available through rc interface
    ###################################################################
    def cmd_help(self):
        """
        returns available list of commands
        """
        return COMMANDS_HELP

    def cmd_input_list(self, *args):
        """
        returns router inputs
        """
        result = ''

        if 'detail' in args:
            tmpl_h = "{port:7}{input:>6}{current:>8} {lastbeat:>10}{audio_left:>8}{audio_right:>8} {lastaudio:>10}\n"
            tmpl   = "{port:7}{input:>6}{current:>8} {lastbeat:10.2f}{audio_left:>8}{audio_right:>8} {lastaudio:>10.2f}\n"
            result += tmpl_h.format(
                    port="port",
                    input="input",
                    current="current",
                    lastbeat="last beat",
                    audio_left="audio l",
                    audio_right="audio r",
                    lastaudio="last audio")
            for input in self.router.get_inputs():
                audio_left, audio_right = input.frame_decoder.get_audio_levels()
                curr_in = self.router.get_current_input()
                result += tmpl.format(
                    port=input.port,
                    input='*' if input.is_available else '-',
                    current='*' if curr_in and curr_in.port == input.port else '-',
                    lastbeat=time.time() - input._last_beat,
                    audio_left=audio_left if audio_left is not None else "?",
                    audio_right=audio_right if audio_right is not None else "?",
                    lastaudio=time.time() - input._last_audio_ok
                )
        else:
            for input in self.router.get_inputs():
                result += '{}\n'.format(input.port)

        return result

    def cmd_output_list(self, *args):
        """
        returns router outputs
        """
        result = ''
        if 'detail' in args:
            for output in self.router.get_outputs():
                result += '{} {}\n'.format(output.output, output.is_connected())
        else:
            for output in self.router.get_outputs():
                result += '{}\n'.format(output.output)
        return result

    def cmd_output_add(self, *args):
        """
        Add input to router
        TODO: validate & handle exceptions
        """
        destination = args[0].strip()
        logger.debug('executor: output.add: {}'.format(destination))

        try:
            self.router.add_output(destination)
        except Exception as e:
            return "Could not add output: {}\n".format(e)
        return '{}\n'.format(destination)

    def cmd_output_remove(self, *args):
        """
        Add input to router
        TODO: validate & handle exceptions
        """
        destination = args[0].strip()
        logger.debug('executor: output.emove: {}'.format(destination))

        try:
            self.router.remove_output(destination)
        except Exception as e:
            return "Could not remove output: {}\n".format(e)
        return '{}\n'.format(destination)

    def cmd_input_current(self):
        """
        returns active router input
        """
        curr_in = self.router.get_current_input()
        if not curr_in:
            return 'none\n'
        return '{}\n'.format(curr_in.port)

    def cmd_input_force(self, *args):
        """
        Force input to given port, or switch back to "auto" mode
        """

        forced_port = args[0].strip()
        available_ports = [i.port for i in self.router.get_inputs()] + ['auto']

        if forced_port == 'auto':
            self.router.set_auto_switch()
            return 'auto\n'
        else:
            try:
                self.router.force_input(int(forced_port))
            except ValueError as e:
                return "Could not force port: {}\n".format(e)
            return '{}\n'.format(forced_port)

    def cmd_input_add(self, *args):
        """
        Add input to router
        TODO: validate & handle exceptions
        """
        port = int(args[0].strip())
        logger.debug('executor: input.add: {}'.format(port))

        try:
            self.router.add_input(port)
        except Exception as e:
            return "Could not add port: {}\n".format(e)
        return '{}\n'.format(port)

    def cmd_input_remove(self, *args):
        """
        Add input to router
        TODO: validate & handle exceptions
        """
        port = int(args[0].strip())
        logger.debug('executor: input.remove: {}'.format(port))
        self.router.remove_input(port)
        return '\n'


    def cmd_reload(self):
        """
        reload router configuration from config-file
        """
        logger.debug('executor: reload configuration')
        self.router.reload_configuration()
        return '\n'



# -*- coding: utf-8 -*-
import time
import logging
import zmq

from zmq.eventloop.zmqstream import ZMQStream
from .frame import ZMQFrameDecoder

logger = logging.getLogger(__name__)

class StreamInput(object):
    """
    Input instance.
    """

    connected = False

    def __init__(self, zmq_ctx, port, delay_settings):
        """
        audio threshold is in range [-90..0] dB
        """
        self.zmq_ctx = zmq_ctx
        self.port = port
        self.delay_settings = delay_settings

        time_now = time.time()
        self._last_beat = time_now - float(self.delay_settings.nodata_failover_delay)
        self._last_audio_ok = time_now - float(self.delay_settings.noaudio_failover_delay)
        self._inhibit_switching_until = time_now

        self.stream = self.bind()
        self.frame_decoder = ZMQFrameDecoder()

    def __str__(self):
        return '<StreamInput: port {}>'.format(self.port)

    def bind(self):
        """
        bind zmq input port
        """
        logger.info('Binding socket on port {} - no-data failover delay/recovery: {}s/{}s - no-audio failover delay/recovery: {}s/{}s'.format(
            self.port,
            self.delay_settings.nodata_failover_delay,
            self.delay_settings.nodata_recover_delay,
            self.delay_settings.noaudio_failover_delay,
            self.delay_settings.noaudio_recover_delay,
        ))
        s = self.zmq_ctx.socket(zmq.SUB)
        s.bind('tcp://*:{port}'.format(port=self.port))
        s.setsockopt(zmq.SUBSCRIBE, b"")
        self.connected = True
        return ZMQStream(s)

    def stop(self):
        logger.debug('Stopping socket on port {}'.format(self.port))
        # close zmq socket
        self.stream.close()
        self.connected = False

    def tick(self, msg):
        """
        triggered on every `on_recv` - used to track input availability.
        """
        time_now = time.time()

        # always load frame so that the audio levels are visible in the
        # telnet interface
        self.frame_decoder.load_frame(msg)

        if self._last_beat + self.delay_settings.nodata_failover_delay < time_now:
            inhibit_switching_until = time_now + self.delay_settings.nodata_recover_delay

            if inhibit_switching_until > self._inhibit_switching_until:
                logger.debug("Input {} inhibit until {} because nodata".format(self.port,
                    time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(inhibit_switching_until))))

            self._inhibit_switching_until = max(self._inhibit_switching_until, inhibit_switching_until)

        if self._last_audio_ok + self.delay_settings.noaudio_failover_delay < time_now:
            inhibit_switching_until = time_now + self.delay_settings.noaudio_recover_delay

            if inhibit_switching_until > self._inhibit_switching_until:
                logger.debug("Input {} inhibit until {} because noaudio".format(self.port,
                    time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(inhibit_switching_until))))

            self._inhibit_switching_until = max(self._inhibit_switching_until, inhibit_switching_until)

        if self.delay_settings.audio_threshold != -90:
            audio_left, audio_right = self.frame_decoder.get_audio_levels()
            has_valid_audio = (audio_left is not None) and \
                              (audio_right is not None) and \
                              (audio_left > self.delay_settings.audio_threshold) and \
                              (audio_right > self.delay_settings.audio_threshold)
            if has_valid_audio:
                self._last_audio_ok = time_now
        else:
            # Assume audio is always ok
            self._last_audio_ok = time_now

        self._last_beat = time_now

    @property
    def is_available(self):
        """
        check if the input instance is 'available':
        "last time ticked less than failover duration"
        """
        time_now = time.time()
        has_recently_received_data = self._last_beat > (time_now - float(self.delay_settings.nodata_failover_delay))
        has_valid_audio = self._last_audio_ok > (time_now - float(self.delay_settings.noaudio_failover_delay))
        is_not_inhibited = self._inhibit_switching_until < time_now

        return has_recently_received_data and has_valid_audio and is_not_inhibited



# -*- coding: utf-8 -*-
import logging

import threading
import zmq
from zmq.utils.monitor import recv_monitor_message

logger = logging.getLogger(__name__)

class StreamOutput(object):
    """
    Output instance. Connects to remote socket and relies the zmq frames.

    Uses zmq-socket-monitor to see if we are connected to the subscriber.
    """
    def __init__(self, zmq_ctx, output):
        self.zmq_ctx = zmq_ctx
        self.output = output
        logger.debug('Connecting output to {}'.format(self.output))
        self.connection = self.zmq_ctx.socket(zmq.PUB)
        self._monitor_thread = threading.Thread(target=StreamOutput._monitor, args=(self,))
        self._monitor_thread.start()
        self.connection.connect(self.output)
        self.connected = True

        self._num_subscribers_lock = threading.Lock()
        self._num_subscribers = 0

    def _monitor(self):
        #return
        # TODO: fox zmq monitoring
        try:
            monitor = self.connection.get_monitor_socket()
            while monitor.poll():
                evt = recv_monitor_message(monitor)
                if evt['event'] == zmq.EVENT_CONNECTED:
                    self._num_subscribers_lock.acquire()
                    self._num_subscribers += 1
                    self._num_subscribers_lock.release()
                elif evt['event'] == zmq.EVENT_DISCONNECTED:
                    self._num_subscribers_lock.acquire()
                    self._num_subscribers -= 1
                    self._num_subscribers_lock.release()
                elif evt['event'] == zmq.EVENT_MONITOR_STOPPED:
                    break
            monitor.close()

        except zmq.ZMQError as e:
            # TODO: how to handle this case? what did went wrong?
            logger.error('Unable to initialize monitor - {}'.format(e))

    def is_connected(self):
        self._num_subscribers_lock.acquire()
        num_sub = self._num_subscribers
        self._num_subscribers_lock.release()
        return num_sub > 0

    def stop(self):
        logger.debug('Stopping output to {}'.format(self.output))
        try:
            self.connection.disable_monitor()
        except Exception as e:
            logger.warning('StreamOutput.disconnect exception - error while disabling monitor: {}'.format(e))
        # close zmq socket
        self.connection.close()
        self.connected = False

    def __str__(self):
        self._num_subscribers_lock.acquire()
        num_sub = self._num_subscribers
        self._num_subscribers_lock.release()
        return '<StreamOutput: {} {}>'.format(self.output, num_sub)

    def send(self, frame):
        """
        passing the zmq frame to the output's connection
        """
        self.connection.send(frame, zmq.NOBLOCK)



# -*- coding: utf-8 -*-
import logging
import zmq
import json

from zmq.eventloop import ioloop
from zmq.eventloop.zmqstream import ZMQStream

from .executor import CommandExecutor

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

logger = logging.getLogger(__name__)

class RoutingMinion(CommandExecutor):

    zmq_receive = None
    zmq_send = None

    def __init__(self, master_addr, router):

        super(RoutingMinion, self).__init__(router)

        self.master_addr = master_addr
        self.zmq_connect_to_master()

        self.router = router
        # self.router.register_state_change(self.state_changed)


    def __str__(self):
        return '<RoutingMinion: {}>'.format(self.master_addr)


    def zmq_connect_to_master(self):
        """
        establishes bidirectional connection to master controller.
        """

        zmq_ct = zmq.Context()
        addr = urlparse(self.master_addr)

        receiver_addr = '{scheme}://{hostname}:{port}'.format(
            scheme=addr.scheme,
            hostname=addr.hostname,
            port=addr.port
        )
        receiver = zmq_ct.socket(zmq.PULL)
        receiver.connect(receiver_addr)
        self.zmq_receive = receiver
        logger.debug('Connecting minion-receiver to {}'.format(receiver_addr))


        sender_addr = '{scheme}://{hostname}:{port}'.format(
            scheme=addr.scheme,
            hostname=addr.hostname,
            port=addr.port + 1
        )
        sender = zmq_ct.socket(zmq.PUB)
        sender.connect(sender_addr)
        self.zmq_send = sender
        logger.debug('Connecting minion-sender to {}'.format(sender_addr))

        #
        stream_receive = ZMQStream(self.zmq_receive)
        stream_receive.on_recv_stream(self.receive)



    def send(self, frame):
        """
        passing the zmq frame to the output's connection
        """
        self.zmq_send.send(frame, zmq.NOBLOCK)

    def receive(self, stream, msg, *args, **kwargs):
        print(msg[0])



    def state_changed(self, state):

        print('**********************')
        print('RoutingMinion - state changed')
        print('**********************')

        #state = b'dsdf'

        self.send(json.dumps(state))


    def start(self):
        ioloop.IOLoop.instance().start()
        while True:
            pass



# -*- coding: utf-8 -*-
import os
import oyaml as yaml

from collections import OrderedDict

from .exceptions import ODRouteConfigException

DEFAULT_CONFIG = {
    'name': None,
    'source_ports': [],
    'destinations': [],
    'nodata_failover_delay': 0.5,
    'nodata_recover_delay': 5,
    'noaudio_failover_delay': 10,
    'noaudio_recover_delay': 10,
    'audio_threshold': -90,
    'telnet': None,
    'socket': None,
    'log_level': 'INFO',
    'log_format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s',
}


def load_config_file(path):
    if not os.path.exists(path):
        raise IOError('Config file does not exist: {}'.format(path))

    with open(path, 'r') as config_file:
        config = yaml.load(config_file)

    if not config:
        raise ODRouteConfigException('Unable to read configuration from file')

    parsed_config = parse_config(config)
    parsed_config['config_file'] = path

    return parsed_config


def parse_config(config):
    parsed_config = OrderedDict()

    defaults = OrderedDict(sorted(DEFAULT_CONFIG.items()))

    for key, value in defaults.items():
        """
        kind of ugly handling. make sure to apply default values if config file has
        keys set to 'none' - like:
            ...
            source_ports:
             - 7000
            destinations:
            nodata_failover_delay: 1.0
            ...
        """
        _value = config.get(key, value)
        if not _value:
            _value = value

        parsed_config[key] = _value

    # handle logging config
    parsed_config['log_level'] = parsed_config['log_level'].upper()

    if parsed_config['name']:
        parsed_config['log_format'] = 'odroute-{} {}'.format(parsed_config['name'], parsed_config['log_format'])

    if parsed_config['socket']:
        parsed_config['socket'] = os.path.abspath(parsed_config['socket'])

    return parsed_config
